@@ -258,6 +258,8 @@ UNRESERVED_KEYWORDS = [
     "TEMPORARY",
     "TERMINATED",
     "TIME",
+    "TIMESTAMP_LTZ",
+    "TIMESTAMP_NTZ",
     "TOUCH",
     "TRANSACTION",
     "TRANSACTIONS",
@@ -271,10 +273,12 @@ UNRESERVED_KEYWORDS = [
     "UNBOUNDED",
     "UNCACHE",
     "UNLOCK",
+    "UNPIVOT",
     "UNSET",
     "UPDATE",
     "USE",
     "VALUES",
+    "VARIANT",
     "VIEW",
     "VIEWS",
     "WRITE",
@@ -311,6 +315,8 @@ UNRESERVED_KEYWORDS = [
     "STREAMING",
     "UPDATES",
     "VIOLATION",
+    "TRACK",
+    "HISTORY",
     # Databricks widget
     "WIDGET",
     "DROPDOWN",
