@@ -2,15 +2,16 @@
 
 import logging
 from dataclasses import dataclass
+from typing import List, Tuple, Type, cast
 
-from typing import List, Type, Tuple, cast
-
-from sqlfluff.core.parser import BaseSegment
-from sqlfluff.core.rules import LintFix
-
+from sqlfluff.core.parser import BaseSegment, RawSegment
+from sqlfluff.core.rules import LintFix, LintResult
 from sqlfluff.utils.reflow.elements import ReflowBlock, ReflowPoint, ReflowSequenceType
-from sqlfluff.utils.reflow.reindent import deduce_line_indent
-
+from sqlfluff.utils.reflow.helpers import (
+    deduce_line_indent,
+    fixes_from_results,
+    pretty_segment_name,
+)
 
 # We're in the utils module, but users will expect reflow
 # logs to appear in the context of rules. Hence it's a subset
@@ -88,6 +89,10 @@ class _RebreakLocation:
             span.strict,
         )
 
+    def pretty_target_name(self) -> str:
+        """Get a nicely formatted name of the target."""
+        return pretty_segment_name(self.target)
+
     def has_templated_newline(self, elements: ReflowSequenceType) -> bool:
         """Is either side a templated newline?
 
@@ -137,6 +142,28 @@ class _RebreakLocation:
         )
 
 
+def first_create_anchor(
+    elem_buff: ReflowSequenceType, loc_range: range
+) -> Tuple[RawSegment, ...]:
+    """Handle the potential case of an empty point with the next point with segments.
+
+    While a reflow element's segments are empty, search for the next
+    available element with segments to anchor new element creation.
+    """
+    # https://github.com/sqlfluff/sqlfluff/issues/4184
+    try:
+        create_anchor = next(
+            elem_buff[i].segments for i in loc_range if elem_buff[i].segments
+        )
+    except StopIteration as exc:  # pragma: no cover
+        # NOTE: We don't test this because we *should* always find
+        # _something_ to anchor the creation on, even if we're
+        # unlucky enough not to find it on the first pass.
+        raise NotImplementedError("Could not find anchor for creation.") from exc
+
+    return create_anchor
+
+
 def identify_rebreak_spans(
     element_buffer: ReflowSequenceType, root_segment: BaseSegment
 ) -> List[_RebreakSpan]:
@@ -158,6 +185,20 @@ def identify_rebreak_spans(
             continue
         # Does the element itself have config? (The easy case)
         if elem.line_position:
+            # We should check whether this is a valid place to break based
+            # on whether it's in a templated tag. If it's not a literal, then skip
+            # it.
+            # TODO: We probably only care if the side of the element that we would
+            # break at (i.e. the start if it's `leading` or the end if it's
+            # `trailing`), but we'll go with the blunt logic for simplicity first.
+            if not elem.segments[0].pos_marker.is_literal():
+                reflow_logger.debug(
+                    "        ! Skipping rebreak span on %s because "
+                    "non-literal location.",
+                    elem.segments[0],
+                )
+                continue
+
             # Blocks should only have one segment so it's easy to pick it.
             spans.append(
                 _RebreakSpan(
@@ -177,11 +218,25 @@ def identify_rebreak_spans(
             if elem.depth_info.stack_positions[key].idx != 0:
                 continue
             # Can we find the end?
-            for end_idx in range(idx, len(element_buffer) - 2):
+            # NOTE: It's safe to look right to the end here rather than up to
+            # -2 because we're going to end up stepping back by two in the
+            # complicated cases.
+            for end_idx in range(idx, len(element_buffer)):
                 end_elem = element_buffer[end_idx]
+                final_idx = None
+
                 if not isinstance(end_elem, ReflowBlock):
                     continue
-                if end_elem.depth_info.stack_positions[key].type in ("end", "solo"):
+                elif key not in end_elem.depth_info.stack_positions:
+                    # If we get here, it means the last block was the end.
+                    # NOTE: This feels a little hacky, but it's because of a limitation
+                    # in detecting the "end" and "solo" markers effectively in larger
+                    # sections.
+                    final_idx = end_idx - 2  # pragma: no cover
+                elif end_elem.depth_info.stack_positions[key].type in ("end", "solo"):
+                    final_idx = end_idx
+
+                if final_idx is not None:
                     # Found the end. Add it to the stack.
                     # We reference the appropriate element from the parent stack.
                     target_depth = elem.depth_info.stack_hashes.index(key)
@@ -192,7 +247,7 @@ def identify_rebreak_spans(
                         _RebreakSpan(
                             target,
                             idx,
-                            end_idx,
+                            final_idx,
                             # NOTE: this isn't pretty but until it needs to be more
                             # complex, this works.
                             elem.line_position_configs[key].split(":")[0],
@@ -200,6 +255,7 @@ def identify_rebreak_spans(
                         )
                     )
                     break
+
             # If we find the start, but not the end, it's not a problem, but
             # we won't be rebreaking this span. This is important so that we
             # don't rebreak part of something without the context of what's
@@ -209,7 +265,7 @@ def identify_rebreak_spans(
 
 def rebreak_sequence(
     elements: ReflowSequenceType, root_segment: BaseSegment
-) -> Tuple[ReflowSequenceType, List[LintFix]]:
+) -> Tuple[ReflowSequenceType, List[LintResult]]:
     """Reflow line breaks within a sequence.
 
     Initially this only _moves_ existing segments
@@ -220,6 +276,7 @@ def rebreak_sequence(
     This intentionally does *not* handle indentation,
     as the existing indents are assumed to be correct.
     """
+    lint_results: List[LintResult] = []
     fixes: List[LintFix] = []
     elem_buff: ReflowSequenceType = elements.copy()
 
@@ -237,11 +294,19 @@ def rebreak_sequence(
     # to handle comments differently. There are two other important points:
     # 1. The next newline outward before code (but passing over comments).
     # 2. The point before the next _code_ segment (ditto comments).
-    locations = [_RebreakLocation.from_span(span, elem_buff) for span in spans]
+    locations: List[_RebreakLocation] = []
+    for span in spans:
+        try:
+            locations.append(_RebreakLocation.from_span(span, elem_buff))
+        # If we try and create a location from an incomplete span (i.e. one
+        # where we're unable to find the next newline effectively), then
+        # we'll get an exception. If we do - skip that one - we won't be
+        # able to effectively work with it even if we could construct it.
+        except UnboundLocalError:
+            pass
 
     # Handle each span:
     for loc in locations:
-
         reflow_logger.debug(
             "Handing Rebreak Span (%r: %s): %r",
             loc.line_position,
@@ -269,10 +334,22 @@ def rebreak_sequence(
             if elem_buff[loc.prev.newline_pt_idx].num_newlines():
                 # We're good. It's already leading.
                 continue
+
+            # Generate the text for any issues.
+            pretty_name = loc.pretty_target_name()
+            if loc.strict:  # pragma: no cover
+                # TODO: The 'strict' option isn't widely tested yet.
+                desc = f"{pretty_name.capitalize()} should always start a new line."
+            else:
+                desc = (
+                    f"Found trailing {pretty_name}. Expected only leading "
+                    "near line breaks."
+                )
+
             # Is it the simple case with no comments between the
             # old and new desired locations and only a single following
             # whitespace?
-            elif (
+            if (
                 loc.next.adj_pt_idx == loc.next.pre_code_pt_idx
                 and elem_buff[loc.next.newline_pt_idx].num_newlines() == 1
             ):
@@ -280,16 +357,17 @@ def rebreak_sequence(
                 # Simple case. No comments.
                 # Strip newlines from the next point. Apply the indent to
                 # the previous point.
-                fixes, prev_point = prev_point.indent_to(
+                new_results, prev_point = prev_point.indent_to(
                     next_point.get_indent() or "", before=loc.target
                 )
-                fixes, next_point = next_point.respace_point(
+                new_results, next_point = next_point.respace_point(
                     cast(ReflowBlock, elem_buff[loc.next.adj_pt_idx - 1]),
                     cast(ReflowBlock, elem_buff[loc.next.adj_pt_idx + 1]),
                     root_segment=root_segment,
-                    fixes=fixes,
+                    lint_results=new_results,
                     strip_newlines=True,
                 )
+
                 # Update the points in the buffer
                 elem_buff[loc.prev.adj_pt_idx] = prev_point
                 elem_buff[loc.next.adj_pt_idx] = next_point
@@ -303,21 +381,28 @@ def rebreak_sequence(
                 # the _preceding_ point.
                 fixes.append(LintFix.delete(loc.target))
                 for seg in elem_buff[loc.prev.adj_pt_idx].segments:
-                    fixes.append(LintFix.delete(seg))
+                    if not seg.is_type("dedent"):
+                        fixes.append(LintFix.delete(seg))
 
                 # We always reinsert after the first point, but respace
                 # the inserted point to ensure it's the right size given
                 # configs.
-                fixes, new_point = ReflowPoint(()).respace_point(
+                new_results, new_point = ReflowPoint(()).respace_point(
                     cast(ReflowBlock, elem_buff[loc.next.adj_pt_idx - 1]),
                     cast(ReflowBlock, elem_buff[loc.next.pre_code_pt_idx + 1]),
                     root_segment=root_segment,
-                    fixes=fixes,
+                    lint_results=[],
                     anchor_on="after",
                 )
+
+                create_anchor = first_create_anchor(
+                    elem_buff,
+                    range(loc.next.pre_code_pt_idx, loc.next.adj_pt_idx - 1, -1),
+                )[-1]
+
                 fixes.append(
                     LintFix.create_after(
-                        elem_buff[loc.next.pre_code_pt_idx].segments[-1],
+                        create_anchor,
                         [loc.target],
                     )
                 )
@@ -336,9 +421,23 @@ def rebreak_sequence(
             if elem_buff[loc.next.newline_pt_idx].num_newlines():
                 # We're good, it's already trailing.
                 continue
+
+            # Generate the text for any issues.
+            pretty_name = loc.pretty_target_name()
+            if loc.strict:  # pragma: no cover
+                # TODO: The 'strict' option isn't widely tested yet.
+                desc = (
+                    f"{pretty_name.capitalize()} should always be at the end of a line."
+                )
+            else:
+                desc = (
+                    f"Found leading {pretty_name}. Expected only trailing "
+                    "near line breaks."
+                )
+
             # Is it the simple case with no comments between the
             # old and new desired locations and only one previous newline?
-            elif (
+            if (
                 loc.prev.adj_pt_idx == loc.prev.pre_code_pt_idx
                 and elem_buff[loc.prev.newline_pt_idx].num_newlines() == 1
             ):
@@ -346,16 +445,17 @@ def rebreak_sequence(
                 # Simple case. No comments.
                 # Strip newlines from the previous point. Apply the indent
                 # to the next point.
-                fixes, next_point = next_point.indent_to(
+                new_results, next_point = next_point.indent_to(
                     prev_point.get_indent() or "", after=loc.target
                 )
-                fixes, prev_point = prev_point.respace_point(
+                new_results, prev_point = prev_point.respace_point(
                     cast(ReflowBlock, elem_buff[loc.prev.adj_pt_idx - 1]),
                     cast(ReflowBlock, elem_buff[loc.prev.adj_pt_idx + 1]),
                     root_segment=root_segment,
-                    fixes=fixes,
+                    lint_results=new_results,
                     strip_newlines=True,
                 )
+
                 # Update the points in the buffer
                 elem_buff[loc.prev.adj_pt_idx] = prev_point
                 elem_buff[loc.next.adj_pt_idx] = next_point
@@ -374,20 +474,46 @@ def rebreak_sequence(
                 # We always reinsert before the first point, but respace
                 # the inserted point to ensure it's the right size given
                 # configs.
-                fixes, new_point = ReflowPoint(()).respace_point(
+                new_results, new_point = ReflowPoint(()).respace_point(
                     cast(ReflowBlock, elem_buff[loc.prev.pre_code_pt_idx - 1]),
                     cast(ReflowBlock, elem_buff[loc.prev.adj_pt_idx + 1]),
                     root_segment=root_segment,
-                    fixes=fixes,
+                    lint_results=[],
                     anchor_on="before",
                 )
-                fixes.append(
-                    LintFix.create_before(
-                        elem_buff[loc.prev.pre_code_pt_idx].segments[0],
-                        [loc.target],
-                    )
+
+                lead_create_anchor = first_create_anchor(
+                    elem_buff, range(loc.prev.pre_code_pt_idx, loc.prev.adj_pt_idx + 1)
                 )
 
+                # Attempt to skip dedent elements on reinsertion. These are typically
+                # found at the end of segments, but we don't want to include the
+                # reinserted segment as part of prior code segment's parent segment.
+                prev_code_anchor = next(
+                    (
+                        prev_code_segment
+                        for prev_code_segment in lead_create_anchor
+                        if not prev_code_segment.is_type("dedent")
+                    ),
+                    None,
+                )
+
+                if prev_code_anchor:
+                    fixes.append(
+                        LintFix.create_before(
+                            prev_code_anchor,
+                            [loc.target],
+                        )
+                    )
+                else:
+                    # All segments were dedents, append to the end instead.
+                    fixes.append(
+                        LintFix.create_after(
+                            lead_create_anchor[-1],
+                            [loc.target],
+                        )
+                    )
+
                 elem_buff = (
                     elem_buff[: loc.prev.pre_code_pt_idx]
                     + [new_point]
@@ -403,14 +529,20 @@ def rebreak_sequence(
             # either leading or trailing and needs to be moved onto its
             # own line.
 
+            # Generate the text for any issues.
+            pretty_name = loc.pretty_target_name()
+            desc = (
+                f"{pretty_name.capitalize()}s should always have a line break "
+                "both before and after."
+            )
+
             # First handle the following newlines first (easy).
             if not elem_buff[loc.next.newline_pt_idx].num_newlines():
                 reflow_logger.debug("  Found missing newline after in alone case")
-                pre_fixes, next_point = next_point.indent_to(
+                new_results, next_point = next_point.indent_to(
                     deduce_line_indent(loc.target.raw_segments[-1], root_segment),
                     after=loc.target,
                 )
-                fixes += pre_fixes
                 # Update the point in the buffer
                 elem_buff[loc.next.adj_pt_idx] = next_point
 
@@ -422,11 +554,10 @@ def rebreak_sequence(
                 # but there isn't an unambiguous way to do this, because we
                 # can't be sure what the comments are referring to.
                 # Given that, we take the simple option.
-                post_fixes, prev_point = prev_point.indent_to(
+                new_results, prev_point = prev_point.indent_to(
                     deduce_line_indent(loc.target.raw_segments[0], root_segment),
                     before=loc.target,
                 )
-                fixes += post_fixes
                 # Update the point in the buffer
                 elem_buff[loc.prev.adj_pt_idx] = prev_point
 
@@ -435,4 +566,14 @@ def rebreak_sequence(
                 f"Unexpected line_position config: {loc.line_position}"
             )
 
-    return elem_buff, fixes
+        # Consolidate results and consume fix buffer
+        lint_results.append(
+            LintResult(
+                loc.target,
+                fixes=fixes_from_results(new_results) + fixes,
+                description=desc,
+            )
+        )
+        fixes = []
+
+    return elem_buff, lint_results
