@@ -1,17 +1,27 @@
 """Dataclasses for reflow work."""
 
-from itertools import chain
 import logging
-from typing import Iterator, List, Optional, Sequence, Tuple, cast, Type
-from sqlfluff.core.config import FluffConfig
+from itertools import chain
+from typing import Iterator, List, Optional, Sequence, Tuple, Type, cast
 
+from sqlfluff.core.config import FluffConfig
 from sqlfluff.core.parser import BaseSegment, RawSegment
-from sqlfluff.core.rules.base import LintFix
+from sqlfluff.core.rules import LintFix, LintResult
 from sqlfluff.utils.reflow.config import ReflowConfig
 from sqlfluff.utils.reflow.depthmap import DepthMap
-
-from sqlfluff.utils.reflow.elements import ReflowBlock, ReflowPoint, ReflowSequenceType
+from sqlfluff.utils.reflow.elements import (
+    ReflowBlock,
+    ReflowPoint,
+    ReflowSequenceType,
+    get_consumed_whitespace,
+)
+from sqlfluff.utils.reflow.helpers import fixes_from_results
 from sqlfluff.utils.reflow.rebreak import rebreak_sequence
+from sqlfluff.utils.reflow.reindent import (
+    construct_single_indent,
+    lint_indent_points,
+    lint_line_length,
+)
 
 # We're in the utils module, but users will expect reflow
 # logs to appear in the context of rules. Hence it's a subset
@@ -55,7 +65,7 @@ class ReflowSequence:
         root_segment: BaseSegment,
         reflow_config: ReflowConfig,
         depth_map: DepthMap,
-        embodied_fixes: Optional[List[LintFix]] = None,
+        lint_results: Optional[List[LintResult]] = None,
     ):
         # First validate integrity
         self._validate_reflow_sequence(elements)
@@ -68,61 +78,35 @@ class ReflowSequence:
         # Alternatively pictured: This is the list of fixes required
         # to generate this sequence. We can build on this as we edit
         # the sequence.
-        self.embodied_fixes: List[LintFix] = embodied_fixes or []
+        # Rather than saving *fixes* directly, we package them into
+        # LintResult objects to make it a little easier to expose them
+        # in the CLI.
+        self.lint_results: List[LintResult] = lint_results or []
 
     def get_fixes(self) -> List[LintFix]:
-        """Get the current fix buffer."""
-        return self.embodied_fixes
+        """Get the current fix buffer.
 
-    def get_partitioned_fixes(
-        self, target: BaseSegment
-    ) -> Tuple[List[LintFix], List[LintFix], List[LintFix]]:
-        """Get the current fix buffer partitioned around a target."""
-        first_target_raw = target.raw_segments[0]
-        last_target_raw = target.raw_segments[-1]
+        We're hydrating them here directly from the LintResult
+        objects, so for more accurate results, consider using
+        .get_results(). This method is particularly useful
+        when consolidating multiple results into one.
+        """
+        return fixes_from_results(self.lint_results)
 
-        assert target.pos_marker
-        pre_fixes = [
-            fix
-            for fix in self.embodied_fixes
-            if fix.anchor.pos_marker
-            and (
-                fix.anchor.pos_marker.working_loc
-                < first_target_raw.pos_marker.working_loc
-                or (
-                    fix.edit_type == "create_before"
-                    and fix.anchor.pos_marker.working_loc
-                    == first_target_raw.pos_marker.working_loc
-                )
-            )
-        ]
-        post_fixes = [
-            fix
-            for fix in self.embodied_fixes
-            if fix.anchor.pos_marker
-            and (
-                fix.anchor.pos_marker.working_loc
-                > last_target_raw.pos_marker.working_loc
-                or (
-                    fix.edit_type == "create_after"
-                    and fix.anchor.pos_marker.working_loc
-                    == last_target_raw.pos_marker.working_loc
-                )
-            )
-        ]
-        # The rest
-        mid_fixes = [
-            fix for fix in self.embodied_fixes if fix not in pre_fixes + post_fixes
-        ]
-        return pre_fixes, mid_fixes, post_fixes
+    def get_results(self) -> List[LintResult]:
+        """Return the current result buffer."""
+        return self.lint_results
 
     def get_raw(self) -> str:
         """Get the current raw representation."""
         return "".join(elem.raw for elem in self.elements)
 
     @staticmethod
-    def _validate_reflow_sequence(elements: ReflowSequenceType):
-        assert elements, "ReflowSequence has empty elements."
+    def _validate_reflow_sequence(elements: ReflowSequenceType) -> None:
+        # An empty set of elements _is_ allowed as an edge case.
+        if not elements:
+            # Return early if so
+            return None
         # Check odds and evens
         OddType = elements[0].__class__
         EvenType = ReflowPoint if OddType is ReflowBlock else ReflowBlock
@@ -135,6 +119,7 @@ class ReflowSequence:
             assert all(
                 isinstance(elem, EvenType) for elem in elements[1::2]
             ), f"Not all even elements are {EvenType.__name__}"
+            return None
         except AssertionError as err:  # pragma: no cover
             for elem in elements:
                 reflow_logger.error("   - %s", elem)
@@ -155,7 +140,12 @@ class ReflowSequence:
         for seg in segments:
             # NOTE: end_of_file is block-like rather than point-like.
             # This is to facilitate better evaluation of the ends of files.
-            if seg.is_type("whitespace", "newline", "indent"):
+            # NOTE: This also allows us to include literal placeholders for
+            # whitespace only strings.
+            if (
+                seg.is_type("whitespace", "newline", "indent")
+                or (get_consumed_whitespace(seg) or "").isspace()
+            ):
                 # Add to the buffer and move on.
                 seg_buff.append(seg)
                 continue
@@ -166,7 +156,7 @@ class ReflowSequence:
             # Add the block, with config info.
             elem_buff.append(
                 ReflowBlock.from_config(
-                    segments=[seg],
+                    segments=(seg,),
                     config=reflow_config,
                     depth_info=depth_map.get_depth_info(seg),
                 )
@@ -334,7 +324,7 @@ class ReflowSequence:
             reflow_config=self.reflow_config,
             depth_map=self.depth_map,
             # Generate the fix to do the removal.
-            embodied_fixes=[LintFix.delete(target)],
+            lint_results=[LintResult(target, [LintFix.delete(target)])],
         )
 
     def insert(
@@ -363,7 +353,7 @@ class ReflowSequence:
         # the target.
         self.depth_map.copy_depth_info(target, insertion)
         new_block = ReflowBlock.from_config(
-            segments=[insertion],
+            segments=(insertion,),
             config=self.reflow_config,
             depth_info=self.depth_map.get_depth_info(target),
         )
@@ -380,7 +370,9 @@ class ReflowSequence:
                 reflow_config=self.reflow_config,
                 depth_map=self.depth_map,
                 # Generate the fix to do the removal.
-                embodied_fixes=[LintFix.create_before(target, [insertion])],
+                lint_results=[
+                    LintResult(target, [LintFix.create_before(target, [insertion])])
+                ],
             )
         elif pos == "after":  # pragma: no cover
             # TODO: This doesn't get coverage - should it even exist?
@@ -394,7 +386,9 @@ class ReflowSequence:
                 reflow_config=self.reflow_config,
                 depth_map=self.depth_map,
                 # Generate the fix to do the removal.
-                embodied_fixes=[LintFix.create_after(target, [insertion])],
+                lint_results=[
+                    LintResult(target, [LintFix.create_after(target, [insertion])])
+                ],
             )
         raise ValueError(
             f"Unexpected value for ReflowSequence.insert(pos): {pos}"
@@ -408,8 +402,6 @@ class ReflowSequence:
         This generates appropriate replacement :obj:`LintFix` objects to direct
         the linter to modify those elements.
         """
-        replace_fix = LintFix.replace(target, edit)
-
         target_raws = target.raw_segments
         assert target_raws
 
@@ -450,7 +442,7 @@ class ReflowSequence:
             root_segment=self.root_segment,
             reflow_config=self.reflow_config,
             depth_map=self.depth_map,
-            embodied_fixes=[replace_fix],
+            lint_results=[LintResult(target, [LintFix.replace(target, edit)])],
         )
 
     def _iter_points_with_constraints(
@@ -487,7 +479,7 @@ class ReflowSequence:
                 most useful for filtering between trailing whitespace
                 and fixes between content on a line.
 
-        **NOTE** this method relies on the embodied fixes being correct
+        **NOTE** this method relies on the embodied results being correct
         so that we can build on them.
         """
         assert filter in (
@@ -496,23 +488,23 @@ class ReflowSequence:
             "inline",
         ), f"Unexpected value for filter: {filter}"
         # Use the embodied fixes as a starting point.
-        fixes = self.embodied_fixes or []
+        lint_results = self.get_results()
         new_elements: ReflowSequenceType = []
         for point, pre, post in self._iter_points_with_constraints():
             # We filter on the elements POST RESPACE. This is to allow
             # strict respacing to reclaim newlines.
-            new_fixes, new_point = point.respace_point(
+            new_lint_results, new_point = point.respace_point(
                 prev_block=pre,
                 next_block=post,
                 root_segment=self.root_segment,
-                fixes=fixes,
+                lint_results=lint_results,
                 strip_newlines=strip_newlines,
             )
             # If filter has been set, optionally unset the returned values.
             if (
                 filter == "inline"
+                # NOTE: We test on the NEW point.
                 if (
-                    # NOTE: We test on the NEW point.
                     any(seg.is_type("newline") for seg in new_point.segments)
                     # Or if it's followed by the end of file
                     or (post and "end_of_file" in post.class_types)
@@ -526,10 +518,7 @@ class ReflowSequence:
                 new_point = point
             # Otherwise apply the new fixes
             else:
-                reflow_logger.debug(
-                    "    Filter %r allows fixes for point: %s", filter, new_fixes
-                )
-                fixes = new_fixes
+                lint_results = new_lint_results
 
             if pre and (not new_elements or new_elements[-1] != pre):
                 new_elements.append(pre)
@@ -541,8 +530,7 @@ class ReflowSequence:
             root_segment=self.root_segment,
             reflow_config=self.reflow_config,
             depth_map=self.depth_map,
-            # Generate the fix to do the removal.
-            embodied_fixes=fixes,
+            lint_results=lint_results,
         )
 
     def rebreak(self) -> "ReflowSequence":
@@ -558,18 +546,81 @@ class ReflowSequence:
             but eventually this method will also handle line
             length considerations too.
         """
-        if self.embodied_fixes:
+        if self.lint_results:
             raise NotImplementedError(  # pragma: no cover
                 "rebreak cannot currently handle pre-existing embodied fixes."
             )
 
         # Delegate to the rebreak algorithm
-        elem_buff, fixes = rebreak_sequence(self.elements, self.root_segment)
+        elem_buff, lint_results = rebreak_sequence(self.elements, self.root_segment)
 
         return ReflowSequence(
             elements=elem_buff,
             root_segment=self.root_segment,
             reflow_config=self.reflow_config,
             depth_map=self.depth_map,
-            embodied_fixes=fixes,
+            lint_results=lint_results,
+        )
+
+    def reindent(self) -> "ReflowSequence":
+        """Reindent lines within a sequence."""
+        if self.lint_results:
+            raise NotImplementedError(  # pragma: no cover
+                "rebreak cannot currently handle pre-existing embodied fixes."
+            )
+
+        single_indent = construct_single_indent(
+            indent_unit=self.reflow_config.indent_unit,
+            tab_space_size=self.reflow_config.tab_space_size,
+        )
+
+        reflow_logger.info("# Evaluating indents.")
+        elements, indent_results = lint_indent_points(
+            self.elements,
+            single_indent=single_indent,
+            skip_indentation_in=self.reflow_config.skip_indentation_in,
+            allow_implicit_indents=self.reflow_config.allow_implicit_indents,
+            ignore_comment_lines=self.reflow_config.ignore_comment_lines,
+        )
+
+        return ReflowSequence(
+            elements=elements,
+            root_segment=self.root_segment,
+            reflow_config=self.reflow_config,
+            depth_map=self.depth_map,
+            lint_results=indent_results,
+        )
+
+    def break_long_lines(self) -> "ReflowSequence":
+        """Rebreak any remaining long lines in a sequence.
+
+        This assumes that reindent() has already been applied.
+        """
+        if self.lint_results:
+            raise NotImplementedError(  # pragma: no cover
+                "break_long_lines cannot currently handle pre-existing "
+                "embodied fixes."
+            )
+
+        single_indent = construct_single_indent(
+            indent_unit=self.reflow_config.indent_unit,
+            tab_space_size=self.reflow_config.tab_space_size,
+        )
+
+        reflow_logger.info("# Evaluating line lengths.")
+        elements, length_results = lint_line_length(
+            self.elements,
+            self.root_segment,
+            single_indent=single_indent,
+            line_length_limit=self.reflow_config.max_line_length,
+            allow_implicit_indents=self.reflow_config.allow_implicit_indents,
+            trailing_comments=self.reflow_config.trailing_comments,
+        )
+
+        return ReflowSequence(
+            elements=elements,
+            root_segment=self.root_segment,
+            reflow_config=self.reflow_config,
+            depth_map=self.depth_map,
+            lint_results=length_results,
         )
